/**
 * A demo of a cloth falling on a cube box, to get started with soft bodies.
 */

#include "demo_skeleton.h"
#include "softbody_mesh.h"

#include <BulletSoftBody/btSoftBody.h>
#include <BulletSoftBody/btSoftBodyHelpers.h>

class ClothOnBox : public DemoSkeleton {
public:
	ClothOnBox();

protected:
	void setupScene() override;
	void renderIteration() override;

private:

	Model mCube;
	btDefaultMotionState *mCubeMotionState = nullptr;

	btSoftBody *mCloth = nullptr;
	SoftbodyMesh mClothMesh;
};

ClothOnBox::ClothOnBox()
{
	mShowPhysicsDebug = true;
	mDebugDrawer.setDebugMode(1);
}

void ClothOnBox::setupScene()
{
	createGroundPlane();

	{
		const float cubeSize = 5.0f;
		Mesh cubeMesh = GenMeshCube(cubeSize, cubeSize, cubeSize);
		mCube = LoadModelFromMesh(cubeMesh);
		btCollisionShape *cubeShape = new btBoxShape(
			{ cubeSize / 2, cubeSize / 2, cubeSize / 2 });
		mCollisionShapes.push_back(cubeShape);

		btTransform cubeTransform;
		cubeTransform.setIdentity();
		cubeTransform.setOrigin({ 0, cubeSize / 2, 0 });
		transformToMatrix(cubeTransform, mCube.transform);

		btScalar mass = 1;
		btVector3 localInertia;
		cubeShape->calculateLocalInertia(mass, localInertia);
		mCubeMotionState = new btDefaultMotionState(cubeTransform);
		btRigidBody::btRigidBodyConstructionInfo rbInfo(
			mass, mCubeMotionState, cubeShape, localInertia);

		btRigidBody *body = new btRigidBody(rbInfo);
		mWorld->addRigidBody(body);
	}

	{
		const float width = 15;
		const float height = 3;
		const float y = 15;
		const int resX = 30;
		const int resY = 2;

		mCloth = btSoftBodyHelpers::CreatePatch(mSoftBodyWorldInfo,
			{ -width / 2, y,  height / 2 }, { width / 2, y,  height / 2 },
			{ -width / 2, y, -height / 2 }, { width / 2, y, -height / 2 },
			resX, resY, 0, true);
		mCloth->getCollisionShape()->setMargin(0.001f);
		mCloth->generateBendingConstraints(2, mCloth->appendMaterial());
		mCloth->setTotalMass(0.1);
		mCloth->m_cfg.piterations = 5;
		mCloth->m_cfg.kDP = 0.005f;
		static_cast<btSoftRigidDynamicsWorld *>(mWorld)->addSoftBody(mCloth);

		mClothMesh.mBody = mCloth;
		mClothMesh.createModel();
	}
}

void ClothOnBox::renderIteration()
{
	btTransform transform;
	mCubeMotionState->getWorldTransform(transform);
	transformToMatrix(transform, mCube.transform);
	DrawModel(mCube, { 0, 0, 0 }, 1.0, YELLOW);

	DrawModel(mClothMesh.updateModel(), { 0, 0, 0 }, 1.0, BLUE);

	DrawGrid(20, 1.0f);
}

int main()
{
	ClothOnBox().run();
	return 0;
}
