#include "demo_skeleton.h"

#include <BulletSoftBody/btSoftBodyRigidBodyCollisionConfiguration.h>

void DemoSkeleton::run()
{
	setupRendering();
	setupPhysics();
	setupScene();
	renderLoop();
	finalizePhysics();
	finalizeRendering();
}

void DemoSkeleton::setupRendering()
{
	const int screenWidth = 1280;
	const int screenHeight = 720;
	InitWindow(screenWidth, screenHeight, "Bullet Demo");

	mCamera.position = { 0.0f, 10.0f, 25.0f };
	mCamera.target = { 0.0f, 0.0f, 0.0f };
	mCamera.up = { 0.0f, 1.0f, 0.0f };
	mCamera.fovy = 45.0f;
	mCamera.type = CAMERA_PERSPECTIVE;

	SetCameraMode(mCamera, CAMERA_THIRD_PERSON);
}

void DemoSkeleton::setupPhysics()
{
	mCollisionConfiguration = new btSoftBodyRigidBodyCollisionConfiguration();
	mDispatcher = new btCollisionDispatcher(mCollisionConfiguration);
	mBroadphase = new btDbvtBroadphase();
	mSolver = new btSequentialImpulseConstraintSolver();
	mWorld = new btSoftRigidDynamicsWorld(
		mDispatcher, mBroadphase, mSolver, mCollisionConfiguration);
	mWorld->setGravity({ 0, -9.81, 0 });

	mSoftBodyWorldInfo.m_broadphase = mBroadphase;
	mSoftBodyWorldInfo.m_dispatcher = mDispatcher;
	mSoftBodyWorldInfo.m_gravity = mWorld->getGravity();
	mSoftBodyWorldInfo.m_sparsesdf.Initialize();

	mWorld->setDebugDrawer(&mDebugDrawer);
}

void DemoSkeleton::finalizePhysics()
{
	cleanObjectsAndShapes();
	delete mWorld;
	mWorld = nullptr;
	delete mSolver;
	mSolver = nullptr;
	delete mBroadphase;
	mBroadphase = nullptr;
	delete mDispatcher;
	mDispatcher = nullptr;
	delete mCollisionConfiguration;
	mCollisionConfiguration = nullptr;
}

void DemoSkeleton::finalizeRendering()
{
	CloseWindow();
}

void DemoSkeleton::renderLoop()
{
	/* Max Bullet substeps.
	Does not make sense to turn it into a method, at the moment. */
	const int maxSubSteps = 10;

	if (!mWorld) {
		return;
	}

	SetTargetFPS(120);

	while (!WindowShouldClose()) {
		mWorld->stepSimulation(GetFrameTime() * mTimeScale, maxSubSteps);

		BeginDrawing();
		ClearBackground(RAYWHITE);
		UpdateCamera(&mCamera);
		BeginMode3D(mCamera);
		renderIteration();
		if (mShowPhysicsDebug) {
			mWorld->debugDrawWorld();
			mDebugDrawer.drawFrame();
		}
		EndMode3D();
		if (mShowFps) {
			DrawFPS(10, 10);
		}
		if (mShowPhysicsDebug) {
			mDebugDrawer.drawTexts(mCamera);
		}
		EndDrawing();
	}
}

void DemoSkeleton::createGroundPlane(float boxSize)
{
	if (!mWorld) {
		return;
	}

	{
		btCollisionShape *groundShape = new btStaticPlaneShape(
			{ mCamera.up.x, mCamera.up.y, mCamera.up.z }, 0);
		mCollisionShapes.push_back(groundShape);

		btTransform groundTransform;
		groundTransform.setIdentity();
		groundTransform.setOrigin({ 0, 0, 0 });

		btScalar mass(0.);
		btVector3 localInertia(0, 0, 0);
		btDefaultMotionState *motionState = new btDefaultMotionState(
			groundTransform);
		btRigidBody::btRigidBodyConstructionInfo rbInfo(
			mass, motionState, groundShape, localInertia);

		btRigidBody* body = new btRigidBody(rbInfo);
		mWorld->addRigidBody(body);
	}

	if (boxSize > 0.0f) {
		btCollisionShape *groundShape = new btBoxShape(
			{ boxSize / 2, 0.5f, boxSize / 2 });
		mCollisionShapes.push_back(groundShape);

		btTransform groundTransform;
		groundTransform.setIdentity();
		groundTransform.setOrigin({ 0, -0.5f, 0 });

		btScalar mass(0.);
		btVector3 localInertia(0, 0, 0);
		btDefaultMotionState *motionState = new btDefaultMotionState(
			groundTransform);
		btRigidBody::btRigidBodyConstructionInfo rbInfo(
			mass, motionState, groundShape, localInertia);

		btRigidBody* body = new btRigidBody(rbInfo);
		mWorld->addRigidBody(body);
	}
}

void DemoSkeleton::cleanObjectsAndShapes()
{
	if (mWorld) {
		btCollisionObjectArray &objects = mWorld->getCollisionObjectArray();
		for (int i = mWorld->getNumCollisionObjects() - 1; i >= 0; i--) {
			btCollisionObject *obj = objects[i];
			btRigidBody *body = btRigidBody::upcast(obj);
			if (body && body->getMotionState()) {
				delete body->getMotionState();
			}
			mWorld->removeCollisionObject(obj);
			delete obj;
		}
	}

	for (int i = 0; i < mCollisionShapes.size(); i++) {
		btCollisionShape *shape = mCollisionShapes[i];
		mCollisionShapes[i] = nullptr;
		delete shape;
	}
	mCollisionShapes.clear();
}
