/**
 * A series of common parts of codes that can be useful to all demos.
 */

#include "ray_bt_debug.h"

#include <raylib.h>

#include <btBulletDynamicsCommon.h>
#include <BulletSoftBody/btSoftBody.h>
#include <BulletSoftBody/btSoftRigidDynamicsWorld.h>

#ifndef demo_skeleton_3dfb8913d1931fd9a7fd832abbcca566
#define demo_skeleton_3dfb8913d1931fd9a7fd832abbcca566

/**
 * Demos should extend this class.
 *
 * It divides the whole demo life in some stages, and each one of them is
 * done in a method, and each one of them can be overriden.
 */
class DemoSkeleton {
public:
	/// This function should be called directly by main
	virtual void run();

protected:
	/// Setup the window and camera
	virtual void setupRendering();

	/// Setup the physics
	virtual void setupPhysics();

	/// Setup objects that will be rendered and their physics
	virtual void setupScene() {}

	/**
	 * The render loop.
	 *
	 * Demos might want to use the renderIteration method insted.
	 * If mWorld is nullptr, the loop will not begin.
	 */
	virtual void renderLoop();

	/// Demo specific drawns on each frame
	virtual void renderIteration() {}

	/// Finalize the physics engine
	virtual void finalizePhysics();

	/// Finalize the rendering engine
	virtual void finalizeRendering();

	/**
	 * A helper that creates a ground plane
	 *
	 * This will create both an infinite plane, that can be used for rigid
	 * bodies, and optionally, a box, that works also with soft bodies.
	 *
	 * \param boxSize The size of the box. If non positive, the box will not be
	 *                created
	 */
	void createGroundPlane(float boxSize = 20.0f);

	/// A helper that deletes collision objects and shapes
	void cleanObjectsAndShapes();

	/// The rendering camera
	Camera3D mCamera;

	/// The Bullet world, needed to step simulations on the render loop
	btDynamicsWorld *mWorld = nullptr;

	/// The object that dispatch calculations for overlapping pairs
	btCollisionDispatcher *mDispatcher = nullptr;

	/// The object that detect aabb-overlapping object pairs
	btBroadphaseInterface *mBroadphase = nullptr;

	/// The object that solves constraints
 	btConstraintSolver *mSolver = nullptr;

	/// The object that configure several aspects of Bullet
	btCollisionConfiguration *mCollisionConfiguration = nullptr;

	/// The array of shapes (to use the finalize physics helper)
	btAlignedObjectArray<btCollisionShape *> mCollisionShapes;

	/// A class that keeps togheter some information about the soft body world
	btSoftBodyWorldInfo mSoftBodyWorldInfo;

	/// A debug drawer for Bullet
	RayBtDebug mDebugDrawer;

	/// Tell wheter to draw physics debug
	bool mShowPhysicsDebug = false;

	/// Tell whether to show the FPS in the rendering
	bool mShowFps = true;

	/// Slow down the simulation (or make it faster)
	float mTimeScale = 1.0;
};

/**
 * Transform a btTransform to a Raylib transformation matrix.
 *
 * Basically, Bullet and Raylib have opposite ways to store matrices, and to
 * pass from one to the other, you need to do transposition.
 *
 * \param t The bullet transformation
 * \param[out] mtx The target matrix
 */
inline void transformToMatrix(const btTransform &t, Matrix &mtx)
{
	/* It's just 16 values... Let's just write them by hand and let the compiler
	do its magic. */
	float values[16];
	t.getOpenGLMatrix(values);
	mtx = {
		values[0], values[4], values[ 8], values[12],
		values[1], values[5], values[ 9], values[13],
		values[2], values[6], values[10], values[14],
		values[3], values[7], values[11], values[15]
	};
}

#endif
