#include "softbody_mesh.h"

#include <BulletSoftBody/btSoftBodyHelpers.h>

#include <fstream>
#include <cstdio>
#include <algorithm>

// Raylib has a macro for that in models.c
const static size_t VBO_SIZE = 7;

// I plan to use tinyply only in this file, so I declare the implementation
#define TINYPLY_IMPLEMENTATION
#include "tinyply.h"

Model &SoftbodyMesh::createModel()
{
	Mesh mesh;
	memset(&mesh, 0, sizeof(mesh));
	setMeshContent(mesh);
	rlLoadMesh(&mesh, true);
	/* Careful! This creates a shallow copy of the mesh!
	Doing this at every frame will result in a memory leak.
	To do so you should use UnloadModel, that does not delete the
	pointers contained in the mesh, so it is okay to use also with
	memory areas that we are managing through the vectors. */
	mModel = LoadModelFromMesh(mesh);

	return mModel;
}

Model &SoftbodyMesh::updateModel()
{
	if (mModel.meshCount && mModel.meshes) {
		// We should always have only one mesh
		Mesh &mesh = mModel.meshes[0];

		setMeshContent(mesh);
		rlUpdateMesh(mesh, 0, mVertices.size() / 3);
		rlUpdateMesh(mesh, 2, mNormals.size() / 3);
		// Actually I think indices do not change during simulation
		rlUpdateMesh(mesh, 6, mIndices.size() / 3);
	}
	return mModel;
}

btSoftBody *SoftbodyMesh::createFromFile(const std::string &filename,
	btSoftBodyWorldInfo &worldInfo, const btTransform &transform)
{
	if (mBody) {
		return mBody;
	}

	{
		std::ifstream in(filename.c_str());
		PlyFile ply;
		ply.parse_header(in);

		std::shared_ptr<PlyData> vertices, normals, faces;
		vertices = ply.request_properties_from_element("vertex", { "x", "y", "z" });
		normals = ply.request_properties_from_element("vertex",
			{ "nx", "ny", "nz" });
		faces = ply.request_properties_from_element("face", { "vertex_indices" },
			3);
		ply.read(in);

		float *data = reinterpret_cast<float *>(vertices->buffer.get());
		mVertices.assign(data, data + vertices->count * 3);
		for (size_t i = 0, j = 0; i < vertices->count; i++, j += 3) {
			btVector3 oldVec(mVertices[j], mVertices[j + 1], mVertices[j + 2]);
			btVector3 newVec = transform(oldVec);
			mVertices[j] = newVec.x();
			mVertices[j + 1] = newVec.y();
			mVertices[j + 2] = newVec.z();
		}

		if (normals) {
			btMatrix3x3 normT = transform.getBasis().inverse().transpose();
			data = reinterpret_cast<float *>(normals->buffer.get());
			mNormals.assign(data, data + normals->count * 3);
			for (size_t i = 0, j = 0; i < normals->count; i++, j += 3) {
				btVector3 oldVec(mNormals[j], mNormals[j + 1], mNormals[j + 2]);
				btVector3 newVec = oldVec.dot3(normT[0], normT[1], normT[2]);
				mNormals[j] = newVec.x();
				mNormals[j + 1] = newVec.y();
				mNormals[j + 2] = newVec.z();
			}
		}

		if (faces->t == tinyply::Type::UINT16) {
			unsigned short *fdata =
				reinterpret_cast<unsigned short *>(faces->buffer.get());
			mIndices.assign(fdata, fdata + faces->count * 3);
		} else if (faces->t == tinyply::Type::UINT32) {
			unsigned int *fdata =
				reinterpret_cast<unsigned int *>(faces->buffer.get());
			std::vector<unsigned int> fInt(fdata, fdata + faces->count * 3);
			mIndices.assign(fInt.begin(), fInt.end());
			std::cout << *(std::max_element(fInt.begin(), fInt.end())) << std::endl;
		}
	}

	std::vector<int> intIndices(mIndices.begin(), mIndices.end());
	mBody = btSoftBodyHelpers::CreateFromTriMesh(worldInfo, mVertices.data(),
		intIndices.data(), intIndices.size() / 3);
	std::cout << mBody->m_faces.size() << " " << intIndices.size() << std::endl;
	if (mNormals.size()) {
		auto &nodes = mBody->m_nodes;
		for (int i = 0, j = 0; i < nodes.size(); i++) {
			nodes[i].m_n.setX(mNormals[j++]);
			nodes[i].m_n.setY(mNormals[j++]);
			nodes[i].m_n.setZ(mNormals[j++]);
		}
	}
	return mBody;
}

void SoftbodyMesh::setMeshContent(Mesh &mesh)
{
	if (!mBody) {
		return;
	}

	if (mVbo.size() < VBO_SIZE) {
		mVbo.resize(VBO_SIZE);
	}

	const auto &nodes = mBody->m_nodes;
	mesh.vertexCount = nodes.size();
	mVertices.resize(mesh.vertexCount * 3);
	mNormals.resize(mVertices.size());
	for (int i = 0, j = 0; i < nodes.size(); i++) {
		mVertices[j] = nodes[i].m_x.x();
		mNormals[j++] = nodes[i].m_n.x();
		mVertices[j] = nodes[i].m_x.y();
		mNormals[j++] = nodes[i].m_n.y();
		mVertices[j] = nodes[i].m_x.z();
		mNormals[j++] = nodes[i].m_n.z();
	}

	const auto *n0 = &nodes[0];
	const auto &faces = mBody->m_faces;
	mesh.triangleCount = faces.size();
	mIndices.resize(mesh.triangleCount * 3);
	for (int i = 0, j = 0; i < faces.size(); i++) {
		mIndices[j++] = faces[i].m_n[0] - n0;
		mIndices[j++] = faces[i].m_n[1] - n0;
		mIndices[j++] = faces[i].m_n[2] - n0;
	}

	mesh.vertices = mVertices.data();
	mesh.normals = mNormals.data();
	mesh.indices = mIndices.data();
	mesh.vboId = mVbo.data();
}
